package matt.implementation;

import matt.adapters.MessageSource;
import matt.adapters.MessageTarget;
import matt.protocol.Participant;
import matt.interfaces.Deactivatable;

public class Controller
  implements
    Runnable,
    Deactivatable {

  final Thread reader;
  final Thread writer;
  final Deactivatable threads;

  private class CompositeDeactivatables
    implements
      Deactivatable {

    final Deactivatable source;
    final Deactivatable target;

    CompositeDeactivatables(
      final Deactivatable source,
      final Deactivatable target) {
      this.source = source;
      this.target = target;
    }

    public void deactivate() {
      source.deactivate();
      target.deactivate();
    }

    public void dispose() {
      deactivate();
    }
  }

  private class ReaderThread
    implements
      Runnable,
      Deactivatable {

    final Participant client;
    final MessageSource source;
    boolean active = true;

    public ReaderThread(
      final Participant client,
      final MessageSource source) {
      this.client = client;
      this.source = source;
    }

    public void run() {
      while (active) {
        if (source.hasMessage()) {
          // source may not throw if it has a message
          final String message = source.getMessage();
          try {
            client.sendMessage(message);
          }
          catch (final Exception e) {
            source.bubbleException(e);
          }
        }
      }
    }

    public void deactivate() {
      active = false;
    }

    public void dispose() {
      deactivate();
    }
  }

  private class WriterThread
    implements
      Runnable,
      Deactivatable {

    final Participant client;
    final MessageTarget target;
    boolean active = true;

    public WriterThread(
      final Participant client,
      final MessageTarget target) {
      this.client = client;
      this.target = target;
    }

    public void run() {
      while (active) {
        getMessage();
        try {
          Thread.sleep(250);
        }
        catch (InterruptedException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }

    private void getMessage() {
      try {
        if (client.hasMessages()) {
          // relayMessage cannot throw
          relayMessage(client.receiveMessage());
        }
      }
      catch (Exception e) {
        // exceptions here cannot be caused by target
        target.bubbleException(e);
      }
    }

    private void relayMessage(
      final String message) {
      try {
        // do not bubble target exceptions
        target.putMessage(message);
      }
      catch (final Exception e) {
        // TODO
      }
    }

    public void deactivate() {
      active = false;
    }

    public void dispose() {
      deactivate();
    }
  }

  public Controller(
    final Participant client,
    final MessageSource source,
    final MessageTarget target) {
    ReaderThread r = new ReaderThread(client, source);
    WriterThread w = new WriterThread(client, target);
    reader = new Thread(r);
    writer = new Thread(w);
    threads = new CompositeDeactivatables(r, w);
  }

  public void run() {
    reader.start();
    writer.start();
  }

  public void deactivate() {
    threads.dispose();
  }

  public void dispose() {
    deactivate();
  }
}
