package matt.implementation;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import matt.interfaces.Deactivatable;
import matt.interfaces.Disposable;

import org.apache.log4j.Logger;

public class NoSpewLogger
  implements
    Disposable {

  private final Logger log;
  private final NoSprew noSpew;
  private final Thread loggingThread;
  private final SimpleDateFormat frater = new SimpleDateFormat("HH:mm:ss.SSS");
  private final Map<String, LogMessage> info = new HashMap<String, LogMessage>();
  private final Map<String, LogMessage> error = new HashMap<String, LogMessage>();

  private class LogMessage {

    private final String message;
    private Date date;
    private long backlog;

    public LogMessage(
      final String message) {
      this.message = message;
      date = new Date();
      backlog = 1;
    }

    public void update(
      final LogMessage message) {
      if (backlog == 0) {
        date = message.date;
      }
      backlog += message.backlog;
    }

    public String toString() {
      if (backlog != 0) {
        final StringBuilder buffer = new StringBuilder();
        buffer.append(message).append(" [calls(").append(backlog);
        buffer.append(") since ").append(frater.format(date)).append(']');
        backlog = 0;
        return buffer.toString();
      }
      return null;
    }
  }

  private class NoSprew
    implements
      Runnable,
      Deactivatable {

    private boolean active = true;
    private final Map<String, LogMessage> info;
    private final Map<String, LogMessage> error;

    public NoSprew(
      final Map<String, LogMessage> info,
      final Map<String, LogMessage> error) {
      this.info = info;
      this.error = error;
    }

    private void printMessages(
      final Map<String, LogMessage> messages) {
      synchronized (messages) {
        List<String> removal = new LinkedList<String>();
        for (final LogMessage msg : messages.values()) {
          final String str = msg.toString();
          if (str != null) {
            if (messages == error) {
              log.error(str);
            }
            else {
              log.info(str);
            }
          }
          else {
            removal.add(msg.message);
          }
        }
        removeMessages(removal, error);
      }
    }

    private void removeMessages(
      final Collection<String> dead,
      final Map<String, LogMessage> log) {
      for (final String msg : dead) {
        log.remove(msg);
      }
      dead.clear();
    }

    public void run() {
      while (active) {
        printMessages(error);
        printMessages(info);
        try {
          Thread.sleep(2000);
        }
        catch (final InterruptedException e) {
          // TODO
        }
      }
    }

    public void deactivate() {
      active = false;
    }

    public void dispose() {
      deactivate();
    }
  }

  public NoSpewLogger(
    final Logger logger) {
    log = logger;
    noSpew = new NoSprew(info, error);
    loggingThread = new Thread(noSpew);
    loggingThread.start();
  }

  private void log(
    final String str,
    final LogMessage msg,
    final Map<String, LogMessage> container) {

    synchronized (container) {
      if (container.containsKey(str)) {
        container.get(str).update(msg);
      }
      else {
        container.put(str, msg);
      }
    }
  }

  public void info(
    final String str) {
    log(str, new LogMessage(str), info);
  }

  public void error(
    final String str) {
    log(str, new LogMessage(str), error);
  }

  public void dispose() {
    noSpew.dispose();
  }
}
