package matt.protocol;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;

import matt.MattAlreadyListeningException;
import matt.MattListenException;
import matt.MattNoMessageException;
import matt.MattNotActiveException;
import matt.MattNotListeningException;
import matt.implementation.NoSpewLogger;
import matt.interfaces.Deactivatable;
import matt.interfaces.Disposable;
import matt.messaging.EventListenerPrx;
import matt.messaging.MailboxPrx;
import matt.roles.ParticipantPrx;
import Ice.Current;

public class Mediator {

  static private final Logger log = Logger.getLogger(Mediator.class);
  static private final NoSpewLogger noSpew = new NoSpewLogger(log);
  private final Runtime runtime;
  private final BlockingQueue<ParticipantPrx> queue = new LinkedBlockingQueue<ParticipantPrx>();
  private final matt.roles.Mediator mediator = new Impl(queue);
  private final PairedMailboxFactory mailboxFactory = new PairedMailboxFactory();
  private final MailboxAssigner mailboxAssigner = new MailboxAssigner(queue);
  private final Thread assignerThread = new Thread(mailboxAssigner);

  private class Impl
    extends
      matt.roles.Mediator {

    private final BlockingQueue<ParticipantPrx> queue;

    Impl(
      final BlockingQueue<ParticipantPrx> queue2) {
      this.queue = queue2;
    }

    public void registerParticipant(
      final ParticipantPrx participant,
      final Current context) {
      boolean queued = false;
      while (queued == false) {
        try {
          noSpew.info("received register request from "
              + participant.ice_toString());
          queue.put(participant);
          queued = true;
        }
        catch (InterruptedException e) {
          // TODO
        }
      }
    }
  }

  private class Mailbox
    extends
      matt.messaging.Mailbox {

    private boolean isActive = true;
    private final BlockingQueue<String> inbox;
    private final BlockingQueue<String> outbox;

    public Mailbox(
      final BlockingQueue<String> inbox,
      final BlockingQueue<String> outbox) {
      this.inbox = inbox;
      this.outbox = outbox;
    }

    public boolean active(
      final Current current) {
      return isActive;
    }

    public void addListener(
      EventListenerPrx arg0,
      final Current current)
      throws MattAlreadyListeningException,
        MattListenException,
        MattNotActiveException {
      // TODO Auto-generated method stub
      assertActive();
      throw new MattListenException("Not Implemented");

    }

    public void deactivate(
      final Current current) {
      noSpew.info("Deactivating for " + current.con._toString().replace('\n', ','));
      isActive = false;
    }

    public boolean empty(
      final Current current) {
      noSpew.info("Empty check for " + current.con._toString().replace('\n', ','));
      return inbox.size() == 0;
    }

    private void assertNotEmpty()
      throws MattNoMessageException {
      if (empty()) {
        throw new MattNoMessageException("There are no messages to read");
      }
    }

    private void assertActive()
      throws MattNotActiveException {
      if (active() == false) {
        throw new MattNotActiveException("The Mailbox is not active");
      }
    }

    public String read(
      final Current current)
      throws MattNoMessageException,
        MattNotActiveException {
      assertActive();
      assertNotEmpty();
      try {
        noSpew.info("reading for " + current.con._toString().replace('\n', ','));
        return inbox.take();
      }
      catch (InterruptedException e) {
        // TODO
        final OutputStream stream = new ByteArrayOutputStream();
        final PrintStream writer = new PrintStream(stream);
        e.printStackTrace(writer);
        final String exception = writer.toString();
        noSpew.error("read: " + exception);
        throw new MattNoMessageException(exception);
      }
    }

    public void removeListener(
      final Current current)
      throws MattListenException,
        MattNotActiveException,
        MattNotListeningException {
      // TODO Auto-generated method stub
      assertActive();
      throw new MattListenException("Not Implemented");
    }

    public void write(
      String message,
      final Current current)
      throws MattNotActiveException {
      // TODO Auto-generated method stub
      assertActive();
      while (true) {
        try {
          noSpew.info("writting for " + current.con._toString().replace('\n', ','));
          outbox.put(message);
          break;
        }
        catch (InterruptedException e) {
          final OutputStream stream = new ByteArrayOutputStream();
          final PrintStream writer = new PrintStream(stream);
          e.printStackTrace(writer);
          final String exception = writer.toString();
          noSpew.info("write: " + exception);
        }
      }
    }
  }

  public class PairedMailbox
    implements
      Disposable,
      Deactivatable {

    private final MailboxPrx first;
    private final MailboxPrx second;

    public PairedMailbox(
      final MailboxPrx first,
      final MailboxPrx second) {
      this.first = first;
      this.second = second;
    }

    public void dispose() {
      deactivate();
      first.ice_getConnection().getAdapter().remove(first.ice_getIdentity());
      second.ice_getConnection().getAdapter().remove(second.ice_getIdentity());
    }

    public void deactivate() {
      first.deactivate();
      second.deactivate();
    }
  }

  private class PairedMailboxFactory {

    private volatile ParticipantPrx firstParticipant = null;
    private volatile MailboxPrx firstMailbox = null;
    private volatile ParticipantPrx secondParticipant = null;
    private volatile MailboxPrx secondMailbox = null;

    public PairedMailboxFactory() {
      // TODO
    }

    public void assignMailbox(
      final ParticipantPrx participant) {
      synchronized (this) {
        if (firstParticipant == null) {
          noSpew.info("Got first participant " + participant.ice_toString());
          firstParticipant = participant;
          while (firstMailbox == null) {
            try {
              this.wait();
            }
            catch (InterruptedException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
          }
        }
        else {
          noSpew.info("Got first participant " + participant.ice_toString());
          secondParticipant = participant;
          BlockingQueue<String> firstQueue = new LinkedBlockingQueue<String>();
          BlockingQueue<String> secondQueue = new LinkedBlockingQueue<String>();
          firstMailbox = runtime.createMailboxPrx(new Mailbox(
            secondQueue,
            firstQueue));
          secondMailbox = runtime.createMailboxPrx(new Mailbox(
            firstQueue,
            secondQueue));
          PairedMailbox paired = new PairedMailbox(firstMailbox, secondMailbox);
          this.notifyAll();
          participant.setMailbox(secondMailbox);
          // TODO something with the paired mailboxes
        }
        noSpew.info("Assigning first participant mailbox " + firstMailbox.ice_toString());
        firstParticipant.setMailbox(firstMailbox);
        noSpew.info("Assigning second participant mailbox " + secondMailbox.ice_toString());
        secondParticipant.setMailbox(secondMailbox);

        firstParticipant = null;
        firstMailbox = null;
        secondParticipant = null;
        secondMailbox = null;
      }
    }
  }

  private class MailboxAssigner
    implements
      Runnable {

    // TODO add deactivate
    private volatile boolean active = true;
    private final BlockingQueue<ParticipantPrx> queue;

    public MailboxAssigner(
      BlockingQueue<ParticipantPrx> queue) {
      this.queue = queue;
    }

    public void run() {
      ParticipantPrx firstParticipant = null;
      while (active) {
        try {
          final ParticipantPrx tmp = queue.take();
          if (firstParticipant == null) {
            noSpew.info("Got first participant " + tmp.ice_toString());
            firstParticipant = tmp;
          }
          else {
            noSpew.info("Got second participant " + tmp.ice_toString());
            final ParticipantPrx secondParticipant = tmp;
            final BlockingQueue<String> firstQueue = new LinkedBlockingQueue<String>();
            final BlockingQueue<String> secondQueue = new LinkedBlockingQueue<String>();
            final MailboxPrx firstMailbox = runtime
              .createMailboxPrx(new Mailbox(secondQueue, firstQueue));
            final MailboxPrx secondMailbox = runtime
              .createMailboxPrx(new Mailbox(firstQueue, secondQueue));
            final PairedMailbox paired = new PairedMailbox(
              firstMailbox,
              secondMailbox);
            noSpew.info("Assigning first participant mailbox " + firstMailbox.ice_toString());
            firstParticipant.setMailbox(firstMailbox);
            noSpew.info("Assigning second participant mailbox " + secondMailbox.ice_toString());
            secondParticipant.setMailbox(secondMailbox);
            firstParticipant = null;
          }
        }
        catch (final InterruptedException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }

  }

  public Mediator(
    final Runtime runtime) {
    this.runtime = runtime;
    assignerThread.start();
  }

  public matt.roles.Mediator getMediator() {
    return mediator;
  }
}
