
#ifndef MATT_COMPONENTS_ICE_20080309_1545
#define MATT_COMPONENTS_ICE_20080309_1545

/**
 * This Ice file outlines the distributed components for a MATT session.
 *
 * TIG participants are required to implement the roles::Participant class,
 * which is expected to implement the setMailbox method.
 * 
 * Participants then have the option to either implementing an eventListener
 * which requires an onEvent method to be implemented or to devise some method
 * by which they poll a mailbox for new messages.
 * 
 * Generated code for C++, C#, Java, Python and Visual Basic will be provided
 * and arrangements to support other languages can be made.
 * 
 * Sample code will also be provided for creating proxies to a Mediator.
 * 
 * Usage:
 *
 * 1. Participant instances connect to a Mediator service
 *
 * 2. Once a sufficient number of Participants have register the Mediator will 
 *    assign mailboxes to each
 *
 * 3. Once a Participant has a mailbox assigned they may begin:
 *    Read/Write Operations on that Mailbox until the Mailbox is closed.
 *    They may register an EventListener while the Mailbix is active.
 *
 *    NOTE: A Mailbox may close for the following reasons:
 *      a. The allotted Time To Live (TTL) has been reached
 *      b. A registered EventListener fails to be responsive(*)
 *
 * 4. Once the Mailbox is closed/deactivated it will not be reopened.
 *
 * 5. Participants will have to re-register with a Mediator service in order
 *    to obtain a new Mailbox 
 *
 * Please contact Marc Allen for more information: matt@1bdi.co.uk 
 */ 

module matt {

  /**
   * Exception base type 
   */
  exception MattException {

    /**
     * Description of the exception
     */
    string description;
  };

  /**
   * This exception is thrown when an RMI occurs on an inactive entity 
   */
  exception MattNotActiveException extends MattException {
  };

  /**
   * This exception is thrown when an RMI fails because there are not pending messages
   */
  exception MattNoMessageException extends MattException {
  };

  /**
   * Base Listen Exception - something went trying to (add|remove) a listener
   */
  exception MattListenException extends MattException {
  };

  /**
   * Exception thrown because there is already an EventListener registered
   */
  exception MattAlreadyListeningException extends MattListenException {
  };

  /**
   * Exception thrown because there was no previous EventListener to unregister 
   */
  exception MattNotListeningException extends MattListenException {
  };

  /**
   * This module handles communications between TIG participants
   */
  module messaging { 

    /**
     * Valid communication events
     */
    enum EventType {
      EventTypeMessage,
      EventTypeOpened,
      EventTypeClosed
    }; 

    /**
     * Participants may choose to have message pushed to them
     */
    class EventListener {
      /**
       * Callback member function describing the event that triggered
       */
      void onEvent(EventType event, string message);
    };

    /**
     * Facilitates read/write operations on messages
     * The mailbox can be used in either a pull or push mode, 
     * i.e. clients of the mailbox  may register an event_listener
     * which will push new messages as they are received. 
     * Alternatively clients may choose to poll the mailbox for 
     * new messages.
     */
    class Mailbox {

      /**
       * If the mailbox has pending messages this return false
       */
      bool empty();

      /**
       * Returns true if mailbox is still in service 
       */
      bool active();

      /**
       * Deactivates the Mailbox - participants should never call this
       */
      void deactivate();

      /**
       * Returns the first unread message in the mailbox
       * Will throw an exception if (empty() == true)
       */
      string read() 
        throws 
          MattNoMessageException, 
          MattNotActiveException;

      /**
       * Posts a message via the mailbox
       * May throw an exception if (active() == false) 
       */
      void write(string message) 
        throws 
          MattNotActiveException;

      /**
       * Registers an event_listener with the mailbox
       * 
       * The mailbox will trigger the on_event message of the 
       * listener each time a new message is received with an
       * event_type of "message".
       *
       * If the mailbox is closed gracefully the listener will 
       * also be informed with a "closed" event message.
       * 
       * A mailbox will support no more than one event_listener
       * per client.
       *
       * If calling listener->on_event(...) fails the mailbox will
       * assume the client is no longer available and the mailbox will
       * close. 
       */
      void addListener(EventListener * listener) 
        throws 
          MattAlreadyListeningException, 
          MattListenException, 
          MattNotActiveException;

      /**
       * Unregisters an EventListener
       */
      void removeListener() 
        throws 
          MattNotListeningException, 
          MattListenException, 
          MattNotActiveException; 
    };

  };

  /**
   * There are two roles in TIG, this module encapsulates these
   */
  module roles {

    /**
     * Participants communicate with one-another via their mailbox
     * which is assigned to them by the mediator once they have registered.
     */
    class Participant {
      void setMailbox(matt::messaging::Mailbox * mb);
    };

    /**
     * Assigns mailboxes to registered clients.
     */
    class Mediator {
      void registerParticipant(Participant * p);
    };
  
  };
};
#endif/*MATT_COMPONENTS_ICE_20080309_1545*/

